# © SleepECG developers
#
# License: BSD (3-clause)

"""Plotting functions."""

from __future__ import annotations

from itertools import cycle
from typing import TYPE_CHECKING, Optional

import numpy as np

if TYPE_CHECKING:
    import matplotlib.pyplot as plt

from .io.sleep_readers import SleepRecord, SleepStage
from .utils import _STAGE_INTS, _STAGE_NAMES, _merge_sleep_stages, _time_to_sec


def plot_ecg(
    ecg: np.ndarray,
    fs: float,
    title: Optional[str] = None,
    **kwargs: np.ndarray,
) -> tuple["plt.Figure", "plt.Axes"]:
    """
    Plot ECG time series with optional markers.

    Parameters
    ----------
    ecg : np.ndarray
        ECG signal.
    fs : float
        Sampling frequency of the ECG signal in Hz.
    title : str, optional
        Title of the plot.
    **kwargs : np.ndarray
        Positions of annotations (i.e. heartbeats) in samples. If more than one marker
        sequence is given, the keywords will be used as labels in the plot legend.

    Returns
    -------
    fig : matplotlib.figure.Figure
        The figure.
    ax : matplotlib.axes.Axes
        The axes in the figure.

    Examples
    --------
    Plot ECG without any annotations:

    >>> plot(ecg, fs)

    Plot ECG with one set of annotations and a title:

    >>> plot(ecg, fs, title="ECG", markers=annotations)

    Plot ECG with two sets of annotations:

    >>> plot(ecg, fs, marker1=annotations, marker2=heartbeats)

    The last example will create two annotation series, the first one labeled `marker1` with
    positions given by `annotations`, and the second one labeled `marker2` with positions
    given by `heartbeats`.
    """
    import matplotlib.pyplot as plt
    from matplotlib.cm import get_cmap

    t = np.arange(0, len(ecg) / fs, 1 / fs)
    fig, ax = plt.subplots()
    ax.plot(t, ecg, color="dimgray", label=None)
    ax.set_xlabel("Time (s)")

    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    colors = cycle(get_cmap("tab10").colors)
    markers = cycle(("*", "o", "s", "D", "v", "<", ">", "^", "X", "p"))
    for label, pos, color, marker in zip(kwargs.keys(), kwargs.values(), colors, markers):
        ax.plot(
            t[pos],
            ecg[pos],
            marker=marker,
            markeredgecolor=color,
            markerfacecolor="None",
            linestyle="",
            label=label,
        )
    if len(kwargs) >= 2:  # add legend if there are at least two sequences of annotations
        plt.legend()
    if title is not None:
        ax.set_title(title)

    return fig, ax


def plot_hypnogram(
    record: SleepRecord,
    stages_pred: np.ndarray,
    stages_mode: str,
    stages_pred_duration: int = 30,
    merge_annotations: bool = False,
    show_bpm: bool = False,
) -> tuple["plt.Figure", list["plt.Axes"]]:
    """
    Plot a hypnogram for a single record.

    Annotated sleep stages are included in the plot if available in `record`. If
    `stages_pred` contains probabilities, they are shown in an additional subplot.

    Parameters
    ----------
    record : SleepRecord
        A single record (i.e. night).
    stages_pred : np.ndarray
        The predicted stages, either as a 1D array of integers or a 2D array of
        probabilties.
    stages_mode : str
        Identifier of the grouping mode. Can be any of `'wake-sleep'`, `'wake-rem-nrem'`,
        `'wake-rem-light-n3'`, or `'wake-rem-n1-n2-n3'`.
    stages_pred_duration : int, optional
        Duration of the predicted sleep stages in seconds, by default `30`.
    merge_annotations : bool, optional
        If `True`, merge annotations according to `stages_mode`, otherwise plot original
        annotations. By default `False`.
    show_bpm : bool, optional
        If `True`, include a subplot of the heart rate in bpm. This can be helpful to find
        bad signal quality intervals, by default `False`.

    Returns
    -------
    fig : matplotlib.figure.Figure
        The figure.
    ax : array of matplotlib.axes.Axes
        The axes in the figure.
    """
    import matplotlib.dates as mdates
    import matplotlib.pyplot as plt

    if record.sleep_stage_duration is None:
        raise ValueError(f"sleep_stage_duration not available for record {record.id}")

    stages_pred_probs = None
    num_subplots = 1
    if stages_pred.ndim == 2:
        num_subplots += 1
        stages_pred_probs = stages_pred
        stages_pred = stages_pred_probs.argmax(1)
    if record.sleep_stages is not None:
        num_subplots += 1
    if show_bpm:
        num_subplots += 1

    if record.recording_start_time is None:
        start_time = 0
    else:
        start_time = _time_to_sec(record.recording_start_time)

    fig, ax = plt.subplots(num_subplots, sharex=True, figsize=(7, 4))

    # predicted stages
    t_stages_pred = np.arange(len(stages_pred)) * stages_pred_duration + start_time
    t_stages_pred = t_stages_pred.astype("datetime64[s]")
    stages_pred = stages_pred.astype(float)
    stages_pred[stages_pred == SleepStage.UNDEFINED] = np.nan
    ax[0].plot(t_stages_pred, stages_pred)
    ax[0].set_yticks(_STAGE_INTS[stages_mode])
    ax[0].set_yticklabels(_STAGE_NAMES[stages_mode])
    ax[0].set_ylabel("predicted")
    ax[0].yaxis.tick_right()

    row = 1

    # predicted stage probabilities
    if stages_pred_probs is not None:
        ax[row].stackplot(
            t_stages_pred,
            stages_pred_probs[:, 1:].T,
            labels=_STAGE_NAMES[stages_mode],
        )
        ax[row].set_ylabel("probabilities")
        legend_handles, legend_labels = ax[row].get_legend_handles_labels()
        ax[row].legend(legend_handles[::-1], legend_labels[::-1], loc=(1.01, 0))
        ax[row].set_ylim(0, 1)
        ax[row].set_yticks([])
        row += 1

    # annotated stages
    if record.sleep_stages is not None:
        stages_true = record.sleep_stages
        t_stages_true = (
            np.arange(len(stages_true)) * record.sleep_stage_duration + start_time
        )
        t_stages_true = t_stages_true.astype("datetime64[s]")
        if merge_annotations:
            stages_true = _merge_sleep_stages([stages_true], stages_mode)[0]
            stages_mode_true = stages_mode
        else:
            stages_mode_true = "wake-rem-n1-n2-n3"
        stages_true = stages_true.astype(float)
        stages_true[stages_true == SleepStage.UNDEFINED] = np.nan

        ax[row].plot(t_stages_true, stages_true)
        ax[row].set_yticks(_STAGE_INTS[stages_mode_true])
        ax[row].set_yticklabels(_STAGE_NAMES[stages_mode_true])
        ax[row].set_ylabel("annotated")
        ax[row].yaxis.tick_right()

        row += 1

    # heartrate
    if show_bpm:
        if record.heartbeat_times is None:
            raise ValueError(f"heartbeat_times not available for record {record.id}")
        t_ecg = (record.heartbeat_times[1:] + start_time).astype("datetime64[s]")
        ax[row].plot(t_ecg, 60 / np.diff(record.heartbeat_times))
        ax[row].set_ylabel("beats per minute")
        ax[row].yaxis.tick_right()

    # x axis ticks and label
    ax[-1].xaxis.set_major_formatter(mdates.DateFormatter("%H"))
    if record.recording_start_time is None:
        ax[-1].set_xlabel("time since recording start in hours")
    else:
        ax[-1].set_xlabel("time of day in hours")
    ax[-1].set_xlim(t_stages_pred[0], t_stages_pred[-1])

    return fig, ax


def _plot_confusion_matrix(
    confmat: np.ndarray,
    stage_names: list[str],
) -> tuple["plt.Figure", "plt.Axes"]:
    """
    Create a labeled plot of a confusion matrix.

    Parameters
    ----------
    confmat : np.ndarray
        A confusion matrix, as returned by `confusion_matrix()`.
    stage_names : list[str]
        Class labels which are used as tick labels.

    Returns
    -------
    fig : matplotlib.figure.Figure
        The figure.
    ax : matplotlib.axes.Axes
        The axes in the figure.
    """
    import matplotlib.pyplot as plt

    classes = np.arange(len(confmat))

    fig, ax = plt.subplots()
    ax.imshow(confmat, cmap="Blues", vmin=0, vmax=confmat[1:, 1:].max())
    for i in range(len(stage_names)):
        for j in range(len(stage_names)):
            ax.text(j, i, f"{confmat[i, j]}", ha="center", va="center", color="k")

    ax.set_ylabel("Annotated Stage")
    ax.set_xlabel("Predicted Stage")
    ax.set_xticks(classes)
    ax.set_yticks(classes)
    ax.set_xticklabels(stage_names)
    ax.set_yticklabels(stage_names)
    ax.xaxis.set_label_position("top")
    ax.xaxis.tick_top()

    return fig, ax



# © SleepECG developers
#
# License: BSD (3-clause)

"""Tests to make sure examples don't crash."""

import fnmatch
import runpy
from pathlib import Path

import matplotlib.pyplot as plt
import pytest

EXCLUDE = [
    "*/benchmark/*",
    "*/classifiers/*",
    "*/try_ws_gru_mesa.py",
]

examples_dir = (Path(__file__).parent / "../../examples").resolve()

example_files = {str(f) for f in examples_dir.rglob("*.py")}
for pattern in EXCLUDE:
    example_files -= set(fnmatch.filter(example_files, pattern))


@pytest.mark.parametrize("script", example_files)
def test_example(script, monkeypatch):
    """Run all examples to make sure they don't crash."""
    # Keep matplotlib from showing figures
    monkeypatch.setattr(plt, "show", lambda: None)

    runpy.run_path(script)



# © SleepECG developers
#
# License: BSD (3-clause)

"""Utility functions."""

from __future__ import annotations

import datetime
import warnings
from typing import Any, Callable, Iterable, TypeVar

import numpy as np

from .io.sleep_readers import SleepStage

# required to propagate the return type annotation through _parallel
_Returnable = TypeVar("_Returnable")


def _parallel(
    n_jobs: int,
    function: Callable[..., _Returnable],
    iterable: Iterable,
    *args: Any,
    **kwargs: Any,
) -> list[_Returnable]:
    """
    Apply a function to each element in an iterable in parallel.

    This uses joblib for parallelism. If the package is not available, it falls back to a
    pure Python loop.

    Parameters
    ----------
    n_jobs : int
        The number of jobs to run in parallel. If `1` (default), no parallelism is used.
        `-1` means using all processors.
    function : Callable
        The function to call.
    iterable : Iterable
        `function` will be applied to every element in this iterable.
    *args
        Positional arguments to be passed to `function`.
    **kwargs
        Keyword arguments to be passed to `function`.

    Returns
    -------
    list
        A list containing the return values of `function`.

    Warnings
    --------
    Note that in case the `function` is very simple, the cost for spawning workers will make
    the parallel execution slower than the standard execution.

    Examples
    --------
    To parallelize this:

    >>> def root(a, n):
    >>>     return a**(1/n)
    >>> res = [root(x, 3) for x in range(1000)]

    Write this:

    >>> n_jobs = 4
    >>> res = _parallel(n_jobs, root, range(1000), 3)
    """
    try:
        from joblib import Parallel, delayed
    except ImportError:
        if n_jobs != 1:
            warnings.warn("joblib not installed, cannot run in parallel.")
        return [function(x, *args, **kwargs) for x in iterable]

    return Parallel(n_jobs=n_jobs)(delayed(function)(x, *args, **kwargs) for x in iterable)


def _time_to_sec(time: datetime.time) -> int:
    """
    Convert a `datetime.time` to seconds.

    `00:00:00` corresponds to `0` and `23:59:59` to `86399`.

    Parameters
    ----------
    time : datetime.time
        A `datetime.time` to convert.

    Returns
    -------
    int
        Time of day in seconds.
    """
    return time.hour * 3600 + time.minute * 60 + time.second


# Classifiers don't always discriminate between all sleep stages defined by the AASM
# guidelines. This dictionary is used to create a consistent mapping from groups of AASM
# sleep stages (as defined in `SleepStage`) to integers. `SleepStage.UNDEFINED` is always
# `0` and the actual stages' values increase with wakefulness, so they map correctly to the
# y-axis in a hypnogram plot. Gaps between stage values are avoided as non-existing classes
# in a one-hot encoding leads to issues when calculating class weights and losses.

_SLEEP_STAGE_MAPPING = {
    "wake-sleep": {
        SleepStage.WAKE: 2,
        SleepStage.REM: 1,
        SleepStage.N1: 1,
        SleepStage.N2: 1,
        SleepStage.N3: 1,
    },
    "wake-rem-nrem": {
        SleepStage.WAKE: 3,
        SleepStage.REM: 2,
        SleepStage.N1: 1,
        SleepStage.N2: 1,
        SleepStage.N3: 1,
    },
    "wake-rem-light-n3": {
        SleepStage.WAKE: 4,
        SleepStage.REM: 3,
        SleepStage.N1: 2,
        SleepStage.N2: 2,
        SleepStage.N3: 1,
    },
    "wake-rem-n1-n2-n3": {
        SleepStage.WAKE: 5,
        SleepStage.REM: 4,
        SleepStage.N1: 3,
        SleepStage.N2: 2,
        SleepStage.N3: 1,
    },
}

_STAGE_NAMES = {m: m.upper().split("-")[::-1] for m in _SLEEP_STAGE_MAPPING}
_STAGE_INTS = {k: sorted(set(v.values())) for k, v in _SLEEP_STAGE_MAPPING.items()}


def _merge_sleep_stages(stages: list[np.ndarray], stages_mode: str) -> list[np.ndarray]:
    """
    Merge sleep stage labels into groups.

    Parameters
    ----------
    stages : list[np.ndarray]
        A list of 1D arrays containing AASM sleep stages as defined by `SleepStage`, e.g. as
        returned by `extract_features()`.
    stages_mode : str
        Identifier of the grouping mode. Can be any of `'wake-sleep'`, `'wake-rem-nrem'`,
        `'wake-rem-light-n3'`, `'wake-rem-n1-n2-n3'`.

    Returns
    -------
    list[np.ndarray]
        A list of 1D arrays containing merged sleep stages.
    """
    if stages_mode not in _SLEEP_STAGE_MAPPING:
        options = list(_SLEEP_STAGE_MAPPING.keys())
        raise ValueError(f"Invalid stages_mode: {stages_mode}. Possible options: {options}")

    new_stages = []
    for array in stages:
        new_array = np.full_like(array, fill_value=SleepStage.UNDEFINED)
        for source_stage, target_stage in _SLEEP_STAGE_MAPPING[stages_mode].items():
            new_array[array == source_stage] = target_stage
        new_stages.append(new_array)
    return new_stages



# © SleepECG developers
#
# License: BSD (3-clause)

"""Tests for the configuration interface."""

import pytest

import sleepecg.config
from sleepecg import get_config, set_config


@pytest.fixture(autouse=True)
def temp_test_config(tmp_path):
    """Create, use and delete a temporary user config file for testing."""
    # setup
    user_config_path_backup = sleepecg.config._USER_CONFIG_PATH
    sleepecg.config._USER_CONFIG_PATH = tmp_path / "testconfig.yml"

    # execute test
    yield

    # cleanup
    sleepecg.config._USER_CONFIG_PATH = user_config_path_backup


def test_get_set_del_config():
    """Test setting, getting and deleting a setting."""
    assert get_config("data_dir") == "~/.sleepecg/datasets"
    assert get_config("classifiers_dir") == "~/.sleepecg/classifiers"
    set_config(data_dir="some_dir")
    assert get_config("data_dir") == "some_dir"
    assert get_config("classifiers_dir") == "~/.sleepecg/classifiers"
    set_config(data_dir=None)
    assert get_config("data_dir") == "~/.sleepecg/datasets"
    assert get_config("classifiers_dir") == "~/.sleepecg/classifiers"


def test_set_invalid_config():
    """Test trying to set an invalid configuration key."""
    with pytest.raises(ValueError, match="Trying to set invalid config key:"):
        set_config(invalid_key="some_value")


def test_get_invalid_config():
    """Test trying to get an invalid configuration key."""
    with pytest.raises(ValueError, match="Trying to get invalid config key:"):
        get_config("invalid_key")


def test_get_all_config():
    """Test trying to get all configuration values as a dict."""
    all_config = get_config()
    assert isinstance(all_config, dict)


def test_invalid_key_in_user_config_file():
    """Test writing an invalid key to the user config file."""
    with open(sleepecg.config._USER_CONFIG_PATH, "w") as user_config_file:
        user_config_file.write("invalid_key: 3")

    with pytest.raises(ValueError, match=r"Invalid key\(s\) found .* {'invalid_key'}"):
        get_config("data_dir")



# © SleepECG developers
#
# License: BSD (3-clause)

"""Tests for heartbeat detection."""

import sys

import numpy as np
import pytest
from scipy.misc import electrocardiogram
from scipy.signal import resample_poly

from sleepecg import compare_heartbeats, detect_heartbeats

pytestmark = pytest.mark.c_extension
ecg = electrocardiogram()
fs = 360
y_true = detect_heartbeats(ecg, fs)  # 478 true peaks


def f1_score(y_pred, y_true, max_distance=5):
    """Calculate the F1-score."""
    tp, fp, fn = compare_heartbeats(y_pred, y_true, max_distance=max_distance)
    tp, fp, fn = len(tp), len(fp), len(fn)
    f1 = tp / (tp + 0.5 * (fp + fn))
    return f1


def test_resampling():
    """Test the impact of resampling on heartbeat detection."""
    for fs_new in [72, 90, 120, 180, 360, 720, 1080, 3600]:
        res = fs_new / fs
        if fs_new < fs:
            up = 1
            down = 1 / res
        else:
            down = 1
            up = res
        ecg_res = resample_poly(ecg, up, down)
        beats = detect_heartbeats(ecg_res, fs_new)
        beats = np.round(beats / res).astype(int)
        f1 = f1_score(beats, y_true)
        assert f1 > 0.95, "F1-score after resampling to {fs_new} Hz is below 0.90."


def test_rescaling():
    """Test the impact of rescaling on heartbeat detection."""
    for scale in np.logspace(-4, 4, 9):
        beats = detect_heartbeats(ecg * scale, fs)
        f1 = f1_score(beats, y_true)
        assert f1 == 1, "F1-score after rescaling by {scale} is not 1."


def test_flat_data():
    """Test the impact of flat data on heartbeat detection."""
    # Flat data at the beginning
    ecg_pad_beginning = np.pad(ecg, pad_width=(ecg.size, 0), mode="edge")
    pks_beginning = detect_heartbeats(ecg_pad_beginning, fs)
    assert f1_score(pks_beginning - ecg.size, y_true) == 1

    # Flat data at the end
    ecg_pad_end = np.pad(ecg, pad_width=(0, ecg.size), mode="edge")
    pks_end = detect_heartbeats(ecg_pad_end, fs)
    assert f1_score(pks_end, y_true) == 1

    # Flat data in the middle
    ecg_pad_middle = np.hstack((ecg_pad_end, ecg))
    pks_middle = detect_heartbeats(ecg_pad_middle, fs)
    assert abs(2 * len(y_true) - len(pks_middle)) < 50

    # Entire recording is flat
    with pytest.raises(ValueError):
        detect_heartbeats(np.ones(10000), fs)

    # ecg is empty
    with pytest.raises(ValueError):
        detect_heartbeats([], fs)


def test_squared_moving_integration_args():
    """Test squared moving window integration argument parsing."""
    from sleepecg._heartbeat_detection import _squared_moving_integration

    x = np.array([0, 0, 1, 1, 1, 1, 2, 2, 3, 3, 2, 2, 1, 1, 0, 0, 0, 0, 0, 0])
    window_length = 10

    _squared_moving_integration(x, window_length)
    _squared_moving_integration(x=x, window_length=window_length)
    _squared_moving_integration(x, window_length=window_length)
    _squared_moving_integration(window_length=window_length, x=x)


@pytest.mark.parametrize("window_length", [1, 4, 5, 20])
def test_squared_moving_integration(window_length):
    """Test squared moving window integration calculation."""
    from sleepecg._heartbeat_detection import _squared_moving_integration

    x = np.array([0, 0, 1, 1, 1, 1, 2, 2, 3, 3, 2, 2, 1, 1, 0, 0, 0, 0, 0, 0])

    ref = np.convolve(x**2, np.ones(window_length), mode="same")
    res = _squared_moving_integration(x, window_length)
    assert np.allclose(ref, res)


@pytest.mark.parametrize(
    ["x", "window_length"],
    [
        ("thisisastring", 10),
        (np.arange(20), 15.0),
        (np.arange(20), 15.5),
    ],
)
def test_squared_moving_integration_typechecks(x, window_length):
    """Test squared moving window integration typechecks."""
    from sleepecg._heartbeat_detection import _squared_moving_integration

    with pytest.raises(TypeError):
        _squared_moving_integration(x, window_length)


@pytest.mark.parametrize(
    ["x", "window_length"],
    [
        (np.arange(10), -5),
        (np.arange(10), 0),
        (np.arange(10), 21),
        (np.array(5), 3),
        (np.zeros((5, 5)), 3),
    ],
)
def test_squared_moving_integration_valuechecks(x, window_length):
    """Test squared moving window integration valuechecks."""
    from sleepecg._heartbeat_detection import _squared_moving_integration

    with pytest.raises(ValueError):
        _squared_moving_integration(x, window_length)


def test_thresholding_args():
    """Test thresholding argument parsing."""
    from sleepecg._heartbeat_detection import _thresholding

    filtered_ecg = np.arange(100)
    integrated_ecg = np.arange(100)
    fs = 10

    _thresholding(filtered_ecg, integrated_ecg, fs)
    _thresholding(filtered_ecg, integrated_ecg, fs=fs)
    _thresholding(filtered_ecg, integrated_ecg=integrated_ecg, fs=fs)
    _thresholding(filtered_ecg=filtered_ecg, integrated_ecg=integrated_ecg, fs=fs)


@pytest.mark.parametrize(
    ["filtered_ecg", "integrated_ecg", "fs"],
    [
        ("thisisastring", np.arange(5), 10),
        (np.arange(5), "thisisastring", 10),
        (np.arange(5), np.arange(5), "thisisastring"),
    ],
)
def test_thresholding_typechecks(filtered_ecg, integrated_ecg, fs):
    """Test thresholding typechecks."""
    from sleepecg._heartbeat_detection import _thresholding

    filtered_ecg_refcount = sys.getrefcount(filtered_ecg)
    integrated_ecg_refcount = sys.getrefcount(integrated_ecg)

    with pytest.raises(TypeError):
        _thresholding(filtered_ecg, integrated_ecg, fs)

    assert sys.getrefcount(filtered_ecg) == filtered_ecg_refcount
    assert sys.getrefcount(integrated_ecg) == integrated_ecg_refcount


@pytest.mark.parametrize(
    ["filtered_ecg", "integrated_ecg", "fs"],
    [
        (np.zeros((5, 5)), np.arange(5), 10),
        (np.array(5), np.arange(5), 10),
        (np.arange(5), np.zeros((5, 5)), 10),
        (np.arange(5), np.array(5), 10),
        (np.arange(5), np.arange(5), -5),
        (np.arange(5), np.arange(5), 0),
    ],
)
def test_thresholding_valuechecks(filtered_ecg, integrated_ecg, fs):
    """Test thresholding valuechecks."""
    from sleepecg._heartbeat_detection import _thresholding

    filtered_ecg_refcount = sys.getrefcount(filtered_ecg)
    integrated_ecg_refcount = sys.getrefcount(integrated_ecg)

    with pytest.raises(ValueError):
        _thresholding(filtered_ecg, integrated_ecg, fs)

    assert sys.getrefcount(filtered_ecg) == filtered_ecg_refcount
    assert sys.getrefcount(integrated_ecg) == integrated_ecg_refcount



# %%
from tensorflow.keras import layers, models
from tqdm import tqdm

from sleepecg import (
    evaluate,
    extract_features,
    load_classifier,
    prepare_data_keras,
    print_class_balance,
    read_mesa,
    read_shhs,
    save_classifier,
    set_nsrr_token,
)

# %% Read data and extract features
set_nsrr_token("your-token-here")
records = list(read_mesa())

feature_extraction_params = {
    "lookback": 120,
    "lookforward": 150,
    "feature_selection": [
        "hrv-time",
        "hrv-frequency",
        "recording_start_time",
        "age",
        "gender",
    ],
    "min_rri": 0.3,
    "max_rri": 2,
    "max_nans": 0.5,
}

features_train, stages_train, feature_ids = extract_features(
    tqdm(records),
    **feature_extraction_params,
    n_jobs=-2,
)

# %% Merge sleep stages, pad and mask data as preparation for keras NN
stages_mode = "wake-rem-nrem"

features_train_pad, stages_train_pad, sample_weight = prepare_data_keras(
    features_train,
    stages_train,
    stages_mode,
)
print_class_balance(stages_train_pad, stages_mode)

# %% Define and train model
model = models.Sequential(
    [
        layers.Input((None, features_train_pad.shape[2])),
        layers.Masking(-1),
        layers.BatchNormalization(),
        layers.Dense(64),
        layers.ReLU(),
        layers.Bidirectional(layers.GRU(8, return_sequences=True)),
        layers.Bidirectional(layers.GRU(8, return_sequences=True)),
        layers.Dense(stages_train_pad.shape[-1], activation="softmax"),
    ]
)

model.compile(
    optimizer="rmsprop",
    loss="categorical_crossentropy",
    metrics=["accuracy"],
)
model.build()
model.summary()

# %% Train model
model.fit(
    features_train_pad,
    stages_train_pad,
    epochs=25,
    sample_weight=sample_weight,
)

# %% Store classifier
save_classifier(
    name="wrn-gru-mesa-weighted",
    model=model,
    stages_mode=stages_mode,
    feature_extraction_params=feature_extraction_params,
    mask_value=-1,
    classifiers_dir="./classifiers",
)

# %% Load classifier from disk for validation
clf = load_classifier("wrn-gru-mesa-weighted", "./classifiers")

# %% Read data and extract features
shhs = list(read_shhs())

features_test, stages_test, feature_ids = extract_features(
    tqdm(shhs),
    **clf.feature_extraction_params,
    n_jobs=-2,
)

# %% Predict & evaluate
features_test_pad, stages_test_pad, _ = prepare_data_keras(
    features_test,
    stages_test,
    clf.stages_mode,
)
y_pred = clf.model.predict(features_test_pad)
evaluate(stages_test_pad, y_pred, clf.stages_mode)



# © SleepECG developers
#
# License: BSD (3-clause)

"""Utilities for runtime and detection quality benchmarks."""

from __future__ import annotations

import time
from typing import Any, Iterator

import numpy as np
import sleepecg
from sleepecg.io.ecg_readers import ECGRecord


class HeartpyWarning(Warning):
    pass


def reader_dispatch(db_slug: str, data_dir: str) -> Iterator[ECGRecord]:
    """
    Read ECG records from mitdb, ltdb or gudb.

    Parameters
    ----------
    db_slug : str
        Short identifier of a dataset, e.g. `'mitdb'`.
    data_dir : str
        Directory where all datasets are stored.

    Yields
    ------
    ECGRecord
        Each element in the generator is of type `ECGRecord` and contains the ECG signal
        (`.ecg`), sampling frequency (`.fs`), annotated beat indices (`.annotations`),
        `.lead`, and `.id`.
    """
    readers = {
        "gudb": sleepecg.read_gudb,
        "ltdb": sleepecg.read_ltdb,
        "mitdb": sleepecg.read_mitdb,
    }
    if db_slug not in readers:
        raise ValueError(f"Invalid db_slug: {db_slug}")
    yield from readers[db_slug](data_dir=data_dir)


def detector_dispatch(ecg: np.ndarray, fs: float, detector: str) -> np.ndarray:
    """
    Provide a common interface for different heartbeat detectors.

    Parameters
    ----------
    ecg : np.ndarray
        ECG signal.
    fs : float
        Sampling frequency of the ECG signal in Hz.
    detector : str
        String identifier of the detector to be used.

    Returns
    -------
    np.ndarray
        Indices of detected heartbeats.
    """
    if detector == "mne":
        import mne

        detection = mne.preprocessing.ecg.qrs_detector(fs, ecg, verbose=False)
    elif detector == "wfdb-xqrs":
        import wfdb.processing

        detection = wfdb.processing.xqrs_detect(ecg, fs, verbose=False)
    elif detector == "pyecg-pantompkins":
        import ecgdetectors

        detection = ecgdetectors.Detectors(fs).pan_tompkins_detector(ecg)
    elif detector == "biosppy-hamilton":
        import biosppy

        detection = biosppy.signals.ecg.hamilton_segmenter(ecg, fs)[0]
    elif detector == "heartpy":
        import heartpy
        from heartpy.exceptions import BadSignalWarning

        try:
            wd, _ = heartpy.process(ecg, fs)
        except BadSignalWarning:
            raise HeartpyWarning
        else:
            detection = np.array(wd["peaklist"])[wd["binary_peaklist"].astype(bool)]
    elif detector == "neurokit2-nk":
        import neurokit2

        clean_ecg = neurokit2.ecg.ecg_clean(ecg, int(fs), method="neurokit")
        detection = neurokit2.ecg.ecg_findpeaks(clean_ecg, int(fs), method="neurokit")[
            "ECG_R_Peaks"
        ]
    elif detector == "neurokit2-kalidas2017":
        import neurokit2

        clean_ecg = neurokit2.ecg.ecg_clean(ecg, int(fs), method="kalidas2017")
        detection = neurokit2.ecg.ecg_findpeaks(clean_ecg, int(fs), method="kalidas2017")[
            "ECG_R_Peaks"
        ]
    elif detector == "sleepecg-c":
        detection = sleepecg.detect_heartbeats(ecg, fs, backend="c")
    elif detector == "sleepecg-numba":
        detection = sleepecg.detect_heartbeats(ecg, fs, backend="numba")
    elif detector == "sleepecg-python":
        detection = sleepecg.detect_heartbeats(ecg, fs, backend="python")
    else:
        raise ValueError(f"Unknown QRS detector: {detector}")
    return np.asarray(detection)


def evaluate_single(
    record: ECGRecord,
    detector: str,
    signal_len: int,
    max_distance: float,
    calc_rri_similarity: bool,
) -> dict[str, Any]:
    """
    Evaluate a heartbeat detector on a given annotated ECG record.

    Optionally, similarity measures between detected and annotated RR intervals can be
    calculated. As this requires interpolation, it may take some time for long signals.

    Parameters
    ----------
    record : ECGRecord
        As received from `reader_dispatch`.
    detector : str
        String identifier of the detector to be used.
    signal_len : int
        Length to which the signal should be sliced.
    max_distance : float
        Maximum temporal distance in seconds between detected and annotated beats to count
        as a successful detection.
    calc_rri_similarity : bool
        If `True`, calculate similarity measures between detected and annotated RR intervals
        (computationally expensive for long signals).

    Returns
    -------
    dict[str, Any]
        A dictionary containing evaluation results.
    """
    signal_len_samples = int(signal_len * record.fs * 60)
    ecg = record.ecg[:signal_len_samples]
    annotation = record.annotation[record.annotation < signal_len_samples]
    fs = int(record.fs)

    try:
        start = time.perf_counter()
        detection = detector_dispatch(ecg, fs, detector)
        runtime = time.perf_counter() - start
        TP, FP, FN = sleepecg.compare_heartbeats(
            detection,
            annotation,
            int(max_distance * record.fs),
        )

        if calc_rri_similarity:
            pearsonr, spearmanr, rmse = sleepecg.rri_similarity(detection, annotation)

    except HeartpyWarning:
        runtime = np.nan
        TP = []
        FP = []
        FN = annotation

        if calc_rri_similarity:
            pearsonr = np.nan
            spearmanr = np.nan
            rmse = np.nan

    result = {
        "record_id": record.id,
        "lead": record.lead,
        "fs": record.fs,
        "num_samples": len(ecg),
        "detector": detector,
        "max_distance": max_distance,
        "runtime": runtime,
        "TP": len(TP),
        "FP": len(FP),
        "FN": len(FN),
    }
    if calc_rri_similarity:
        result.update(
            {
                "pearsonr": pearsonr,
                "spearmanr": spearmanr,
                "rmse": rmse,
            }
        )
    return result



# © SleepECG developers
#
# License: BSD (3-clause)

"""Functions for reading datasets containing ECG and beat annotations."""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Iterator, Optional

import numpy as np
from tqdm import tqdm

if TYPE_CHECKING:
    import matplotlib.pyplot as plt

from ..config import get_config
from ..plot import plot_ecg
from .gudb import _GUDB_MD5
from .physionet import _list_physionet, download_physionet
from .utils import _download_file


@dataclass
class ECGRecord:
    """
    Store a single ECG record.

    Attributes
    ----------
    ecg : np.ndarray
        The ECG signal.
    fs : float
        The sampling frequency.
    annotation : np.ndarray
        Indices of annotated heartbeats.
    lead : str, optional
        Which ECG lead the signal was recorded from, by default `None`.
    id : str, optional
        The record ID, by default `None`.
    """

    ecg: np.ndarray
    fs: float
    annotation: np.ndarray
    lead: Optional[str] = None
    id: Optional[str] = None

    def export(self, filename: str | Path) -> None:
        """
        Export ECG record to CSV.

        Parameters
        ----------
        filename : str | pathlib.Path
            File name to write to.
        """
        export_ecg_record(self, filename)

    def plot(self, **kwargs: np.ndarray) -> tuple["plt.Figure", "plt.Axes"]:
        """
        Plot ECG time series with optional markers.

        Parameters
        ----------
        **kwargs : np.ndarray
            Positions of annotations (i.e. heartbeats) in samples. If more than one marker
            sequence is given, the keywords will be used as labels in the plot legend.

        Returns
        -------
        fig : matplotlib.figure.Figure
            The figure.
        ax : matplotlib.axes.Axes
            The axes in the figure.
        """
        return plot_ecg(self.ecg, self.fs, title=self.id, beats=self.annotation, **kwargs)


def export_ecg_record(record: ECGRecord, filename: str | Path) -> None:
    """
    Export ECG record to CSV.

    Parameters
    ----------
    record : ECGRecord
        ECG record to export.
    filename : str | pathlib.Path
        File name to write to.
    """
    filename = Path(filename).with_suffix(".csv")

    rpeaks = np.zeros_like(record.ecg, dtype=int)
    rpeaks[record.annotation] = 1

    np.savetxt(
        filename,
        np.vstack((record.ecg, rpeaks)).T,
        fmt="%.3f,%d",
        header=f"# fs: {record.fs}Hz\necg,rpeak",
        comments="",
    )


def read_ltdb(
    records_pattern: str = "*",
    offline: bool = False,
    data_dir: Optional[str | Path] = None,
) -> Iterator[ECGRecord]:
    """
    Lazily read records from [LTDB](https://physionet.org/content/ltdb/).

    Parameters
    ----------
    records_pattern : str, optional
        Glob-like pattern to select record IDs, by default `'*'`.
    offline : bool, optional
        If `True`, only local files will be used (i.e. no files will be downloaded), by
        default `False`.
    data_dir : str | pathlib.Path, optional
        Directory where all datasets are stored. If `None` (default), the value will be
        taken from the configuration.

    Yields
    ------
    ECGRecord
        Each element in the generator is of type `ECGRecord` and contains the ECG signal
        (`.ecg`), sampling frequency (`.fs`), annotated beat indices (`.annotations`),
        `.lead`, and `.id`.
    """
    if data_dir is None:
        data_dir = get_config("data_dir")
    yield from _read_mitbih("ltdb", records_pattern, offline, data_dir)


def read_mitdb(
    records_pattern: str = "*",
    offline: bool = False,
    data_dir: Optional[str | Path] = None,
) -> Iterator[ECGRecord]:
    """
    Lazily read records from [MITDB](https://physionet.org/content/mitdb/).

    Parameters
    ----------
    records_pattern : str, optional
        Glob-like pattern to select record IDs, by default `'*'`.
    offline : bool, optional
        If `True`, only local files will be used (i.e. no files will be downloaded), by
        default `False`.
    data_dir : str | pathlib.Path, optional
        Directory where all datasets are stored. If `None` (default), the value will be
        taken from the configuration.

    Yields
    ------
    ECGRecord
        Each element in the generator is of type `ECGRecord` and contains the ECG signal
        (`.ecg`), sampling frequency (`.fs`), annotated beat indices (`.annotations`),
        `.lead`, and `.id`.
    """
    if data_dir is None:
        data_dir = get_config("data_dir")
    yield from _read_mitbih("mitdb", records_pattern, offline, data_dir)


def _read_mitbih(
    db_slug: str,
    records_pattern: str,
    offline: bool,
    data_dir: str | Path,
) -> Iterator[ECGRecord]:
    """
    Lazily reads records from MIT-BIH datasets (e.g. MITDB, LTDB).

    Required files are downloaded if not present in `<data_dir>/<db_slug>`.

    Parameters
    ----------
    db_slug : str
        Short identifier of a database, e.g. `'mitdb'`.
    records_pattern : str
        Glob-like pattern to select record IDs.
    offline : bool
        If `True`, only local files will be used (i.e. no files will be downloaded).
    data_dir : str | pathlib.Path
        Directory where all datasets are stored.

    Yields
    ------
    ECGRecord
        Each element in the generator is of type `ECGRecord` and contains the ECG signal
        (`.ecg`), sampling frequency (`.fs`), annotated beat indices (`.annotations`),
        `.lead`, and `.id`.
    """
    import wfdb

    # https://archive.physionet.org/physiobank/database/html/mitdbdir/intro.htm#symbols
    BEAT_ANNOTATION_SYMBOLS = set("NLRAaJSFejE/fQ|")

    data_dir = Path(data_dir).expanduser()

    requested_records = _list_physionet(data_dir, db_slug, pattern=records_pattern)

    if not offline:
        download_physionet(
            db_slug,
            requested_records,
            extensions=[".hea", ".dat", ".atr"],
            data_dir=data_dir,
        )

    for record_id in requested_records:
        record_file = str(data_dir / db_slug / record_id)

        annotations = wfdb.rdann(record_file, "atr")
        beat_indices = []
        for sample, symbol in zip(annotations.sample, annotations.symbol):
            if symbol in BEAT_ANNOTATION_SYMBOLS:
                beat_indices.append(sample)

        record = wfdb.rdrecord(record_file)

        for signal_index, signal_name in enumerate(record.sig_name):
            yield ECGRecord(
                ecg=record.p_signal[:, signal_index].flatten(),
                fs=record.fs,
                annotation=np.array(beat_indices, dtype=int),
                lead=signal_name,
                id=record_id,
            )


def read_gudb(
    offline: bool = False,
    data_dir: Optional[str | Path] = None,
) -> Iterator[ECGRecord]:
    """
    Lazily read records from [GUDB](https://berndporr.github.io/ECG-GUDB/).

    Required files are downloaded if not present in `'<data_dir>/gudb'`.

    Parameters
    ----------
    offline : bool, optional
        If `True`, only local files will be used (i.e. no files will be downloaded), by
        default `False`.
    data_dir : str | pathlib.Path, optional
        Directory where all datasets are stored. If `None` (default), the value will be
        taken from the configuration.

    Yields
    ------
    ECGRecord
        Each element in the generator is of type `ECGRecord` and contains the ECG signal
        (`.ecg`), sampling frequency (`.fs`), annotated beat indices (`.annotations`),
        `.lead`, and `.id`.
    """
    DB_URL = "https://berndporr.github.io/ECG-GUDB/experiment_data"
    EXPERIMENTS = ["sitting", "maths", "walking", "hand_bike", "jogging"]
    FS = 250

    if data_dir is None:
        data_dir = get_config("data_dir")

    db_dir = Path(data_dir).expanduser() / "gudb"

    for subject_id in tqdm(list(range(25)), desc="Reading GUDB"):
        for experiment in EXPERIMENTS:
            experiment_subdir = f"subject_{subject_id:02}/{experiment}"
            if not offline:
                for tsv_filename in (
                    "ECG.tsv",
                    "annotation_cs.tsv",
                    "annotation_cables.tsv",
                ):
                    ecg_file_url = f"{DB_URL}/{experiment_subdir}/{tsv_filename}"
                    target_filepath = db_dir / experiment_subdir / tsv_filename
                    try:
                        checksum = _GUDB_MD5[f"{experiment_subdir}/{tsv_filename}"]
                    except KeyError:
                        pass  # file not available
                    else:
                        _download_file(ecg_file_url, target_filepath, checksum, "md5")
            ecg_data = {
                lead: signal
                for lead, signal in zip(
                    ("chest", "II", "III"),
                    np.loadtxt(
                        db_dir / experiment_subdir / "ECG.tsv",
                        delimiter=" ",  # space-separated (contrary to what .tsv suggests)
                        usecols=(0, 1, 2),
                        unpack=True,
                    ),
                )
            }
            annotations_chest_file = db_dir / experiment_subdir / "annotation_cs.tsv"
            if annotations_chest_file.is_file():
                yield ECGRecord(
                    ecg=ecg_data["chest"].to_numpy(),
                    fs=FS,
                    annotation=np.loadtxt(annotations_chest_file, dtype=np.int32),
                    lead="chest",
                    id=f"{subject_id:02}_{experiment}",
                )
            annotations_chest_file = db_dir / experiment_subdir / "annotation_cables.tsv"
            if annotations_chest_file.is_file():
                annotations = np.loadtxt(annotations_chest_file, dtype=np.int32)
                for lead in ("II", "III"):
                    yield ECGRecord(
                        ecg=ecg_data[lead].to_numpy(),
                        fs=FS,
                        annotation=annotations,
                        lead=lead,
                        id=f"{subject_id:02}_{experiment}",
                    )



# © SleepECG developers
#
# License: BSD (3-clause)

"""Tests for heartbeat detection and detector evaluation."""

import sys

import numpy as np
import pytest

from sleepecg import compare_heartbeats, detect_heartbeats, read_mitdb


def test_compare_heartbeats():
    """Test heartbeat comparison results."""
    detection = np.array([20, 33, 43, 53, 73])
    annotation = np.array([20, 34, 58, 75, 99])
    max_distance = 3

    TP, FP, FN = compare_heartbeats(detection, annotation, max_distance)

    assert np.all(TP == np.array([20, 33, 73]))
    assert np.all(FP == np.array([43, 53]))
    assert np.all(FN == np.array([58, 99]))


@pytest.fixture(scope="session")
def mitdb_234_MLII():
    """Fetch record for detector tests."""
    return next(read_mitdb(records_pattern="234"))


@pytest.mark.parametrize(
    "backend",
    [
        "c",
        pytest.param(
            "numba",
            marks=pytest.mark.skipif(
                sys.version_info >= (3, 11), reason="numba does not support Python 3.11 yet"
            ),
        ),
        "python",
    ],
)
def test_detect_heartbeats(mitdb_234_MLII, backend):
    """Test heartbeat detection on mitdb:234:MLII."""
    record = mitdb_234_MLII
    detection = detect_heartbeats(record.ecg, record.fs, backend=backend)
    TP, FP, FN = compare_heartbeats(detection, record.annotation, int(record.fs / 10))

    # Changes in the heartbeat detector should not lead to worse results!
    assert len(TP) >= 2750
    assert len(FP) <= 3
    assert len(FN) == 0



# © SleepECG developers
#
# License: BSD (3-clause)

"""Plot results of runtime and detection quality benchmarks."""

import sys
from pathlib import Path

import numpy as np
import pandas as pd
import plotly.express as px

if len(sys.argv) != 2:
    print("Usage: python plot_benchmark_results.py <results>.csv")
    exit()

results_filepath = Path(sys.argv[1])
benchmark, db_slug, *_ = results_filepath.stem.split("__")
plot_filepath = results_filepath.with_suffix(".svg")
results = pd.read_csv(results_filepath).sort_values("detector")

if benchmark == "runtime":
    fs = results["fs"][0]
    results["signal_len"] = results["num_samples"] / (results["fs"] * 3600)
    results = results.groupby(["detector", "signal_len"], as_index=False).agg(
        mean_runtime=("runtime", "mean"),
        std_runtime=("runtime", "std"),
        n=("runtime", "count"),
    )
    results["error"] = results["std_runtime"] / np.sqrt(results["n"])

    # order by runtime for longest signal, slowest algorithm first
    maxlen = results["signal_len"].max()
    order = (
        results.query(f"signal_len == {maxlen}")
        .groupby("detector")["mean_runtime"]
        .mean()
        .apply(lambda x: 1 / x)  # reverse order
        .to_dict()
    )
    results = results.sort_values(by=["detector", "signal_len"], key=lambda x: x.map(order))

    # each detector should have the same color in each benchmark
    colors = [px.colors.qualitative.Plotly[i] for i in [0, 1, 7, 5, 8, 4, 2, 3, 9, 6]]

    fig = (
        px.line(
            results,
            x="signal_len",
            y="mean_runtime",
            markers=True,
            color="detector",
            color_discrete_sequence=colors,
            log_y=True,
            labels={
                "signal_len": "Signal length (hours)",
                "mean_runtime": "Mean runtime (s)",
            },
            title=f"Mean detector runtime for {db_slug.upper()}",
            width=1000,
            template="plotly_white",
        )
        .update_yaxes(rangemode="tozero")
        .update_layout(legend_title="")
    )
    fig.write_image(plot_filepath)

elif benchmark == "metrics":
    results["precision"] = results["TP"] / (results["TP"] + results["FP"])
    results["recall"] = results["TP"] / (results["TP"] + results["FN"])
    results["f1"] = 2 / (results["recall"] ** -1 + results["precision"] ** -1)
    fig = (
        px.box(
            results.melt(id_vars=["detector"], value_vars=["precision", "recall", "f1"]),
            color="detector",
            y="value",
            labels={"value": ""},
            facet_col="variable",
            title=f"Metrics for {db_slug.upper()}",
            width=1000,
            template="plotly_white",
        )
        .update_xaxes(range=[-0.4, 0.4])
        .update_yaxes(range=[-0.01, 1.01], tick0=0, dtick=0.1)
        .for_each_annotation(lambda a: a.update(text=a.text.split("=")[1]))
        .update_layout(legend_title="")
    )
    fig.write_image(plot_filepath)

elif benchmark == "rri_similarity":
    fig = (
        px.box(
            results,
            y="pearsonr",
            color="detector",
            labels={"pearsonr": "Correlation coefficient"},
            title=f"Correlation coefficient for RRI timeseries for {db_slug.upper()}",
            width=1000,
            template="plotly_white",
        )
        .update_yaxes(range=[-1.01, 1.01], tick0=-1, dtick=0.25)
        .update_layout(legend_title="")
    )
    fig.write_image(plot_filepath)

else:
    raise ValueError(f"No plotting strategy defined for {results_filepath}.")



# %% Imports
import matplotlib.pyplot as plt
import numpy as np

from sleepecg import compare_heartbeats, detect_heartbeats, read_mitdb

# %% Download and read data, run detector
record = list(read_mitdb(records_pattern="234"))[1]
detection = detect_heartbeats(record.ecg, record.fs)


# %% Evaluation and visualization
TP, FP, FN = compare_heartbeats(detection, record.annotation, int(record.fs / 10))

t = np.arange(len(record.ecg)) / record.fs

fig, ax = plt.subplots(3, sharex=True, figsize=(10, 8))

ax[0].plot(t, record.ecg, color="k", zorder=1, label="ECG")
ax[0].scatter(
    record.annotation / record.fs,
    record.ecg[record.annotation],
    marker="o",
    color="g",
    s=50,
    zorder=2,
    label="annotation",
)
ax[0].set_ylabel("raw signal in mV")

ax[1].eventplot(
    detection / record.fs,
    linelength=0.5,
    linewidth=0.5,
    color="k",
    zorder=1,
    label="detection",
)
ax[1].scatter(
    FN / record.fs,
    np.ones_like(FN),
    marker="x",
    color="r",
    s=70,
    zorder=2,
    label="FN",
)
ax[1].scatter(
    FP / record.fs,
    np.ones_like(FP),
    marker="+",
    color="orange",
    s=70,
    zorder=2,
    label="FP",
)
ax[1].set_yticks([])
ax[1].set_ylabel("heartbeat events")

ax[2].plot(
    detection[1:] / record.fs,
    60 / (np.diff(detection) / record.fs),
    label="heartrate in bpm",
)
ax[2].set_ylabel("beats per minute")
ax[2].set_xlabel("time in seconds")

for ax_ in ax.flat:
    ax_.legend(loc="upper right")
    ax_.grid(axis="x")

fig.suptitle(
    f"Record ID: {record.id}, lead: {record.lead}\n"
    + f"Recall: {len(TP) / (len(TP) + len(FN)):.2%}, "
    + f"Precision: {len(TP) / (len(TP) + len(FP)):.2%}",
)

plt.show()



# © SleepECG developers
#
# License: BSD (3-clause)

"""Functions related to classifier training and evaluation."""

from __future__ import annotations

import shutil
from dataclasses import dataclass
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Optional, Protocol
from zipfile import ZipFile

import numpy as np
import yaml

from .config import get_config
from .feature_extraction import extract_features
from .io.sleep_readers import SleepRecord, SleepStage
from .utils import _STAGE_NAMES, _merge_sleep_stages


def prepare_data_keras(
    features: list[np.ndarray],
    stages: list[np.ndarray],
    stages_mode: str,
    mask_value: int = -1,
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Mask and pad data and calculate sample weights for a Keras model.

    The following steps are performed:

    - Merge sleep stages in `stages` according to `stage_mode`.
    - Set features corresponding to `SleepStage.UNDEFINED` to `mask_value`.
    - Replace `np.nan` and `np.inf` in `features` with `mask_value`.
    - Pad to a common length, where `mask_value` is used for `features` and
      `SleepStage.UNDEFINED` (i.e `0`) is used for stages.
    - One-hot encode stages.
    - Calculate sample weights with class weights taken as `n_samples /
      (n_classes * np.bincount(y))`.

    Parameters
    ----------
    features : list[np.ndarray]
        Each 2D array in this list is a feature matrix of shape `(n_samples, n_features)`
        corresponding to a single record as returned by `extract_features()`.
    stages : list[np.ndarray]
        Each 1D array in this list contains the sleep stages of a single record as returned
        by `extract_features()`.
    stages_mode : str
        Identifier of the grouping mode. Can be any of `'wake-sleep'`, `'wake-rem-nrem'`,
        `'wake-rem-light-n3'`, `'wake-rem-n1-n2-n3'`.
    mask_value : int, optional
        Value used to pad features and replace `np.nan` and `np.inf`, by default `-1`.
        Remember to pass the same value to `layers.Masking` in your model.

    Returns
    -------
    features_padded : np.ndarray
        A 3D array of shape `(n_records, max_n_samples, n_features)`, where `n_records` is
        the length of `features`/`stages` and `max_n_samples` is the maximum number of rows
        of all feature matrices in `features`.
    stages_padded_onehot : np.ndarray
        A 3D array of shape `(n_records, max_n_samples, n_classes+1)`, where `n_classes` is
        the number of classes remaining after merging sleep stages (excluding
        `SleepStage.UNDEFINED`).
    sample_weight : np.ndarray
        A 2D array of shape `(n_records, max_n_samples)`.
    """
    from tensorflow.keras.preprocessing.sequence import pad_sequences
    from tensorflow.keras.utils import to_categorical

    stages_merged = _merge_sleep_stages(stages, stages_mode)
    stages_padded = pad_sequences(stages_merged, value=SleepStage.UNDEFINED)
    stages_padded_onehot = to_categorical(stages_padded)

    features_padded = pad_sequences(features, dtype=float, value=mask_value)
    features_padded[stages_padded == SleepStage.UNDEFINED, :] = mask_value
    features_padded[~np.isfinite(features_padded)] = mask_value

    stage_counts = stages_padded_onehot.sum(0).sum(0)
    # samples corresponding to SleepStage.UNDEFINED are ignored, so their count shouldn't
    # influence the class weights -> slice with [1:]
    class_weight = np.sum(stage_counts[1:]) / stage_counts
    sample_weight = class_weight[stages_padded]

    return features_padded, stages_padded_onehot, sample_weight


def print_class_balance(stages: np.ndarray, stages_mode: Optional[str] = None) -> None:
    """
    Print the number of samples and percentages of each class in `stages`.

    Parameters
    ----------
    stages : np.ndarray
        A 2D array of shape `(n_records, n_samples)` containing integer class labels or a
        3D array of shape `(n_records, n_samples, n_classes)` containing one-hot encoded
        class labels.
    stages_mode : str, optional
        Identifier of the grouping mode. Can be any of `'wake-sleep'`, `'wake-rem-nrem'`,
        `'wake-rem-light-n3'`, `'wake-rem-n1-n2-n3'`. If `None` (default), no class labels
        are printed.
    """
    if stages.ndim == 3:
        stages = stages.argmax(2)

    if stages_mode is not None:
        stage_names = ["UNDEFINED"] + _STAGE_NAMES[stages_mode]
    else:
        stage_names = [str(n) for n in range(6)]

    print("Class balance:")

    unique_stages, counts = np.unique(stages, return_counts=True)
    max_len_counts = len(str(max(counts)))
    max_len_stages = max(len(str(s)) for s in stage_names)
    total_count = counts.sum()
    for stage, count, fraction in zip(unique_stages, counts, counts / total_count):
        print(
            f"    {stage_names[stage]:>{max_len_stages}}: {count:{max_len_counts}} "
            f"({fraction:3.0%})"
        )


def save_classifier(
    name: str,
    model: Any,
    stages_mode: str,
    feature_extraction_params: dict[str, Any],
    mask_value: Optional[int] = None,
    classifiers_dir: Optional[str | Path] = None,
) -> None:
    """
    Save a trained classifier to disk.

    The `model` itself and a `.yml` file containing classifier metadata are stored as
    `<name>.zip` in `classifiers_dir`. Model serialization is performed as suggested by the
    respective package documentation. Currently only Keras models are supported.

    Parameters
    ----------
    name : str
        An identifier which is used as the filename.
    model : Any
        The classification model, should have `fit()` and `predict()` methods.
    stages_mode : str
        Identifier of the grouping mode. Can be any of `'wake-sleep'`, `'wake-rem-nrem'`,
        `'wake-rem-light-n3'`, or `'wake-rem-n1-n2-n3'`.
    feature_extraction_params : dict[str, typing.Any]
        The parameters passed to `extract_features()`, as a dictionary mapping string
        parameter names to values. Should not include `records` and `n_jobs`.
    mask_value : int, optional
        Only required for Keras models, as passed to `prepare_data_keras()` and
        `keras.layers.Masking`, by default `None`.
    classifiers_dir : str | pathlib.Path, optional
        Directory in which the `.zip` file is stored. If `None` (default), the value is
        taken from the configuration.

    See Also
    --------
    load_classifier : Load classifiers saved with this function.
    """
    if classifiers_dir is None:
        classifiers_dir = get_config("classifiers_dir")

    target_file = Path(classifiers_dir).expanduser() / name

    model_type = model.__module__.split(".")[0]
    classifier_info = {
        "model_type": model_type,
        "stages_mode": stages_mode,
        "feature_extraction_params": feature_extraction_params,
    }
    if mask_value is not None:
        classifier_info["mask_value"] = mask_value

    with TemporaryDirectory() as tmpdir:
        with open(f"{tmpdir}/info.yml", "w") as infofile:
            yaml.dump(classifier_info, infofile)

        if model_type == "keras":
            model.save(f"{tmpdir}/classifier")
        else:
            raise ValueError(f"Saving model of type {type(model)} is not supported")

        shutil.make_archive(str(target_file), "zip", tmpdir)


class _Model(Protocol):
    def fit(self, X: np.ndarray, y: np.ndarray) -> None:
        ...

    def predict(self, X: np.ndarray) -> np.ndarray:
        ...


@dataclass
class SleepClassifier:
    """
    Store a sleep classifier model and metadata.

    Attributes
    ----------
    model : _Model
        The classification model, should have `fit` and `predict` methods.
    stages_mode : str
        Identifier of the grouping mode. Can be any of `'wake-sleep'`, `'wake-rem-nrem'`,
        `'wake-rem-light-n3'`, or `'wake-rem-n1-n2-n3'`.
    feature_extraction_params : dict[str, typing.Any]
        The parameters passed to `extract_features()`, as a dictionary mapping string
        parameter names to values. Does not include `records` and `n_jobs`.
    model_type : str
        A string identifying the model type, e.g. `'keras'` or `'sklearn'`. This is used by
        `stage()` to determine how to perform sleep stage predictions.
    mask_value : int, optional
        Only required for models of type `'keras'`, as passed to `prepare_data_keras()` and
        `keras.layers.Masking`, by default `None`.
    source_file : pathlib.Path, optional
        The file from which the classifier was loaded using `load_classifier()`, by default
        `None`.
    """

    model: _Model
    stages_mode: str
    feature_extraction_params: dict[str, Any]
    model_type: str
    mask_value: Optional[int] = None
    source_file: Optional[Path] = None

    def __repr__(self) -> str:
        if self.source_file is not None:
            return (
                f"<SleepClassifier | {self.stages_mode}, {self.model_type}, "
                f"{self.source_file.name}>"
            )
        return f"<SleepClassifier | {self.stages_mode}, {self.model_type}>"

    def __str__(self) -> str:
        features = ", ".join(self.feature_extraction_params["feature_selection"])
        return (
            f"SleepClassifier for {self.stages_mode.upper()}\n"
            f"    features: {features}\n"
            f"    model type: {self.model_type}\n"
            f"    source file: {self.source_file}\n"
        )


def load_classifier(
    name: str,
    classifiers_dir: Optional[str | Path] = None,
) -> SleepClassifier:
    """
    Load a `SleepClassifier` from disk.

    This functions reads `.zip` files saved by `save_classifier()`. Pass `'SleepECG'` as the
    second argument to load a classifier bundled with SleepECG.

    Parameters
    ----------
    name : str
        The identifier of the classifier to load.
    classifiers_dir : str | pathlib.Path, optional
        Directory in which to look for `<name>.zip`. If `None` (default), the value is taken
        from the configuration. If `'SleepECG'`, load classifiers from
        `site-packages/sleepecg/classifiers`.

    Returns
    -------
    SleepClassifier
        Contains the model and metadata required for feature extraction and preprocessing.
        Can be passed to `stage()`.

    See Also
    --------
    list_classifiers : Show information about available classifiers.
    """
    if classifiers_dir == "SleepECG":
        classifiers_dir = Path(__file__).parent / "classifiers"
    elif classifiers_dir is None:
        classifiers_dir = get_config("classifiers_dir")

    soure_file = Path(classifiers_dir).expanduser() / f"{name}.zip"

    with TemporaryDirectory() as tmpdir:
        shutil.unpack_archive(soure_file, tmpdir)

        with open(f"{tmpdir}/info.yml") as infofile:
            classifier_info = yaml.safe_load(infofile)

        if classifier_info["model_type"] == "keras":
            from tensorflow import keras

            classifier = keras.models.load_model(f"{tmpdir}/classifier")

        else:
            raise ValueError(
                f"Loading model of type {classifier_info['model_type']} is not supported"
            )

    return SleepClassifier(
        model=classifier,
        source_file=soure_file,
        **classifier_info,
    )


def list_classifiers(classifiers_dir: Optional[str | Path] = None) -> None:
    """
    List available classifiers.

    Parameters
    ----------
    classifiers_dir : str | pathlib.Path, optional
        Directory in which to look for classifiers. If `None` (default), the value is taken
        from the configuration. If `'SleepECG'`, `site-packages/sleepecg/classifiers` is
        used.

    See Also
    --------
    load_classifier : Load classifiers.
    """
    if classifiers_dir == "SleepECG":
        classifiers_dir = Path(__file__).parent / "classifiers"
        print("Classifiers in SleepECG:")
    elif classifiers_dir is None:
        classifiers_dir = get_config("classifiers_dir")
        print(f"Classifiers in {classifiers_dir}:")
    else:
        print(f"Classifiers in {classifiers_dir}:")

    classifiers_dir = Path(classifiers_dir).expanduser()

    for classifier_filepath in classifiers_dir.glob("*.zip"):
        with ZipFile(classifier_filepath, "r") as zip_file:
            with zip_file.open("info.yml") as infofile:
                classifier_info = yaml.safe_load(infofile)
                features = ", ".join(
                    classifier_info["feature_extraction_params"]["feature_selection"]
                )
                print(
                    f"  {classifier_filepath.stem}\n"
                    f"      stages_mode: {classifier_info['stages_mode'].upper()}\n"
                    f"      model type: {classifier_info['model_type']}\n"
                    f"      features: {features}\n"
                )


def _confusion_matrix(y_true: np.ndarray, y_pred: np.ndarray, N: int) -> np.ndarray:
    """
    Compute confusion matrix.

    Parameters
    ----------
    y_true : np.ndarray
        Ground truth (correct) target values.
    y_pred : np.ndarray
        Estimated targets as returned by a classifier.
    N : int
        Number of unique classes.

    Returns
    -------
    np.ndarray
        Confusion matrix whose i-th row and j-th column entry indicates the number of
        samples with true label being i-th class and predicted label being j-th class.
    """
    return np.bincount(N * y_true + y_pred, minlength=N * N).reshape(N, N)


def _cohen_kappa(confmat: np.ndarray) -> float:
    """
    Compute Cohen's kappa (which measures inter-annotator agreement).

    Implementation modified from `sklearn.metrics.cohen_kappa_score`.

    Parameters
    ----------
    confmat : np.ndarray
        A confusion matrix, as returned by `confusion_matrix()`.

    Returns
    -------
    float
        The kappa statistic, which is a number between -1 and 1. The maximum value means
        complete agreement; zero or lower means chance agreement.
    """
    n_classes = confmat.shape[0]
    sum0 = np.sum(confmat, axis=0)
    sum1 = np.sum(confmat, axis=1)
    expected = np.outer(sum0, sum1) / np.sum(sum0)
    w_mat = 1 - np.eye(n_classes, dtype=int)
    k: float = np.sum(w_mat * confmat) / np.sum(w_mat * expected)
    return 1 - k


def evaluate(
    stages_true: np.ndarray,
    stages_pred: np.ndarray,
    stages_mode: str,
    show_undefined: bool = False,
) -> tuple[np.ndarray, list[str]]:
    """
    Evaluate the performance of a sleep stage classifier.

    Prints overall accuracy, Cohen's kappa, confusion matrix, per-class precision, recall,
    and F1 score.

    Parameters
    ----------
    stages_true : np.ndarray
        The annotated (ground truth) sleep stages as a 2D array of shape
        `(n_records, n_samples)` containing integer class labels, or a 3D array of shape
        `(n_records, n_samples, n_classes)` containing one-hot encoded class labels.
    stages_pred : np.ndarray
        The predicted sleep stages as a 2D array of shape `(n_records, n_samples)`
        containing integer class labels, or a 3D array of shape
        `(n_records, n_samples, n_classes)` containing one-hot encoded class labels.
    stages_mode : str
        Identifier of the grouping mode. Can be any of `'wake-sleep'`, `'wake-rem-nrem'`,
        `'wake-rem-light-n3'`, `'wake-rem-n1-n2-n3'`.
    show_undefined : bool, optional
        If `True`, include `SleepStage.UNDEFINED` (i.e `0`) in the confusion matrix output.
        This can be helpful during debugging. By default `False`.

    Returns
    -------
    conf_mat : np.ndarray
        Confusion matrix.
    stage_names : list[str]
        Sleep stage names.
    """
    stage_names = _STAGE_NAMES[stages_mode]

    if stages_true.ndim == 3:
        stages_true = stages_true.argmax(2)
    if stages_pred.ndim == 3:
        stages_pred = stages_pred.argmax(2)

    confmat_full = _confusion_matrix(
        stages_true.ravel(),
        stages_pred.ravel(),
        len(stage_names) + 1,
    )
    confmat = confmat_full[1:, 1:]

    print(f"Confusion matrix ({stages_mode.upper()}):")
    if show_undefined:
        print(confmat_full)
    else:
        print(confmat)

    kappa = _cohen_kappa(confmat)

    acc = confmat.trace() / confmat.sum()
    tp = np.diag(confmat)
    fp = confmat.sum(1) - tp
    fn = confmat.sum(0) - tp
    precision = tp / (tp + fn)
    recall = tp / (tp + fp)
    f1 = 2 / (recall**-1 + precision**-1)
    support = confmat.sum(1)

    print(f"Accuracy: {acc:.4f}")
    print(f"Cohen's kappa: {kappa:.4f}")
    print("       precision    recall  f1-score    support")
    for i, stage_name in enumerate(stage_names):
        print(
            f"{stage_name:>5}{precision[i]:11.2f}{recall[i]:10.2f}{f1[i]:10.2f}"
            f"{support[i]:11}"
        )
    print(f"{support.sum():47}")

    return confmat_full, stage_names


def stage(
    clf: SleepClassifier,
    record: SleepRecord,
    return_mode: str = "int",
) -> np.ndarray:
    """
    Predict sleep stages for a single record.

    Feature extraction and preprocessing are performed according to the information stored
    in `clf`.

    Parameters
    ----------
    clf : SleepClassifier
        A classifier object as loaded with `load_classifier()`.
    record : SleepRecord
        A single record (i.e. night).
    return_mode : str, optional
        If `'int'`, return the predicted sleep stages as a 1D array of integers. If
        `'prob'`, return a 2D array of probabilities. If `'str'`, return a 1D array of
        strings.

    Returns
    -------
    np.ndarray
        Array of sleep stages. Depending on `return_mode`, this takes different forms.

    Warnings
    --------
    Note that the returned labels depend on `clf.stages_mode`, so they do not necessarily
    follow the stage-to-integer mapping defined in `SleepStage`. See
    [classification](../classification.md) for details.
    """
    return_modes = {"int", "prob", "str"}
    if return_mode not in return_modes:
        raise ValueError(
            f"Invalid return_mode: {return_mode!r}. Possible options: {return_modes}"
        )

    stage_names = ["UNDEFINED"] + _STAGE_NAMES[clf.stages_mode]

    features = extract_features(records=[record], **clf.feature_extraction_params)[0][0]
    if clf.model_type == "keras":
        features[~np.isfinite(features)] = clf.mask_value
        stages_pred_proba: np.ndarray = clf.model.predict(features[np.newaxis, ...])[0]
        stages_pred: np.ndarray = stages_pred_proba.argmax(-1)
    else:
        raise ValueError(f"Staging with model of type {type(clf)} is not supported")

    if return_mode == "prob":
        return stages_pred_proba
    elif return_mode == "str":
        return np.array([stage_names[s] for s in stages_pred])
    return stages_pred



import numpy as np
from setuptools import Extension, setup

setup(
    ext_modules=[
        Extension(
            "sleepecg._heartbeat_detection",
            ["sleepecg/_heartbeat_detection.c"],
            include_dirs=[np.get_include()],
        ),
    ],
)



# © SleepECG developers
#
# License: BSD (3-clause)

"""Functions for downloading and reading datasets."""

from .ecg_readers import ECGRecord, export_ecg_record, read_gudb, read_ltdb, read_mitdb
from .nsrr import download_nsrr, set_nsrr_token
from .physionet import download_physionet
from .sleep_readers import (
    Gender,
    SleepRecord,
    SleepStage,
    SubjectData,
    read_mesa,
    read_shhs,
    read_slpdb,
)



# © SleepECG developers
#
# License: BSD (3-clause)

"""Tests for sleep data reader functions."""

from __future__ import annotations

import datetime
from pathlib import Path

import numpy as np
import scipy.misc
from pyedflib import highlevel

from sleepecg import SleepStage, read_mesa, read_shhs, read_slpdb
from sleepecg.io.sleep_readers import Gender


def _dummy_nsrr_edf(filename: str, hours: float, ecg_channel: str):
    ECG_FS = 360
    ecg_5_min = scipy.misc.electrocardiogram()
    seconds = int(hours * 60 * 60)
    ecg = np.tile(ecg_5_min, int(np.ceil(seconds / 300)))[np.newaxis, : seconds * ECG_FS]
    signal_headers = highlevel.make_signal_headers([ecg_channel], sample_frequency=ECG_FS)
    highlevel.write_edf(filename, ecg, signal_headers)


def _dummy_nsrr_xml(filename: str, hours: float, random_state: int):
    EPOCH_LENGTH = 30
    STAGES = [
        "Wake|0",
        "Stage 1 sleep|1",
        "Stage 2 sleep|2",
        "Stage 3 sleep|3",
        "Stage 4 sleep|4",
        "REM sleep|5",
        "Unscored|9",
    ]

    rng = np.random.default_rng(random_state)

    with open(filename, "w") as xml_file:
        xml_file.write(
            '<?xml version="1.0" encoding="UTF-8" standalone="no"?>\n'
            "<PSGAnnotation>\n"
            f"<EpochLength>{EPOCH_LENGTH}</EpochLength>\n"
            "<ScoredEvents>\n"
            "<ScoredEvent>\n"
            "<EventType/>\n"
            "<EventConcept>Recording Start Time</EventConcept>\n"
            "<ClockTime>01.01.85 20.29.59</ClockTime>\n"
            "</ScoredEvent>\n",
        )
        record_duration = hours * 60 * 60
        start = 0
        while True:
            if start > record_duration:
                break
            epoch_duration = rng.choice(np.arange(4, 21)) * EPOCH_LENGTH
            stage = rng.choice(STAGES)
            xml_file.write(
                "<ScoredEvent>\n"
                "<EventType>Stages|Stages</EventType>\n"
                f"<EventConcept>{stage}</EventConcept>\n"
                f"<Start>{start:.1f}</Start>\n"
                f"<Duration>{epoch_duration:.1f}</Duration>\n"
                "</ScoredEvent>\n",
            )
            start += epoch_duration

        xml_file.write(
            "</ScoredEvents>\n" "</PSGAnnotation>\n",
        )


def _create_dummy_mesa(data_dir: str, durations: list[float], random_state: int = 42):
    DB_SLUG = "mesa"
    ANNOTATION_DIRNAME = "polysomnography/annotations-events-nsrr"
    EDF_DIRNAME = "polysomnography/edfs"
    CSV_DIRNAME = "datasets"

    db_dir = Path(data_dir).expanduser() / DB_SLUG
    annotations_dir = db_dir / ANNOTATION_DIRNAME
    edf_dir = db_dir / EDF_DIRNAME
    csv_dir = db_dir / CSV_DIRNAME

    for directory in (annotations_dir, edf_dir, csv_dir):
        directory.mkdir(parents=True, exist_ok=True)

    for i, hours in enumerate(durations):
        record_id = f"mesa-sleep-{i:04}"
        _dummy_nsrr_edf(f"{edf_dir}/{record_id}.edf", hours, ecg_channel="EKG")
        _dummy_nsrr_xml(f"{annotations_dir}/{record_id}-nsrr.xml", hours, random_state)

    with open(csv_dir / "mesa-sleep-dataset-0.0.0.csv", "w") as csv:
        csv.write("mesaid,examnumber,race1c,gender1,cucmcn1c,sleepage5c\n")
        for i in range(len(durations)):
            csv.write(f"{i},5,0,0,0,77\n")


def _create_dummy_shhs(data_dir: str, durations: list[float], random_state: int = 42):
    DB_SLUG = "shhs"
    ANNOTATION_DIRNAME = "polysomnography/annotations-events-nsrr"
    EDF_DIRNAME = "polysomnography/edfs"
    CSV_DIRNAME = "datasets"

    db_dir = Path(data_dir).expanduser() / DB_SLUG
    annotations_dir = db_dir / ANNOTATION_DIRNAME
    edf_dir = db_dir / EDF_DIRNAME
    csv_dir = db_dir / CSV_DIRNAME

    for directory in (annotations_dir, edf_dir):
        for visit in ("shhs1", "shhs2"):
            (directory / visit).mkdir(parents=True, exist_ok=True)
    csv_dir.mkdir(parents=True, exist_ok=True)

    for i, hours in enumerate(durations):
        for visit in ("shhs1", "shhs2"):
            record_id = f"{visit}/{visit}-20{i:04}"
            _dummy_nsrr_edf(f"{edf_dir}/{record_id}.edf", hours, ecg_channel="ECG")
            _dummy_nsrr_xml(f"{annotations_dir}/{record_id}-nsrr.xml", hours, random_state)

    with open(csv_dir / "shhs1-dataset-0.0.0.csv", "w") as csv:
        csv.write("nsrrid,age_s1,gender,weight\n")
        for i in range(len(durations)):
            csv.write(f"2{i:05},55,1,77\n")
    with open(csv_dir / "shhs2-dataset-0.0.0.csv", "w") as csv:
        csv.write("nsrrid,age_s2,gender,weight\n")
        for i in range(len(durations)):
            csv.write(f"2{i:05},61,2,\n")


def test_read_mesa(tmp_path):
    """Basic sanity checks for records read via read_mesa."""
    durations = [0.1, 0.2]  # hours
    valid_stages = {int(s) for s in SleepStage}

    _create_dummy_mesa(data_dir=tmp_path, durations=durations)
    records = list(read_mesa(data_dir=tmp_path, heartbeats_source="ecg", offline=True))

    assert len(records) == 2

    for rec in records:
        assert rec.sleep_stage_duration == 30
        assert set(rec.sleep_stages) - valid_stages == set()


def test_read_shhs(tmp_path):
    """Basic sanity checks for records read via read_shhs."""
    durations = [0.1, 0.2]  # hours
    valid_stages = {int(s) for s in SleepStage}

    _create_dummy_shhs(data_dir=tmp_path, durations=durations)
    records = list(read_shhs(data_dir=tmp_path, heartbeats_source="ecg", offline=True))

    assert len(records) == 4

    for rec in records:
        assert rec.sleep_stage_duration == 30
        assert set(rec.sleep_stages) - valid_stages == set()


def test_read_slpdb():
    """Basic test for read_slpdb."""
    rec = next(read_slpdb(records_pattern="slp01a"))
    assert rec.sleep_stages.shape == (240,)
    assert rec.sleep_stage_duration == 30
    assert rec.id == "slp01a"
    assert rec.recording_start_time == datetime.time(23, 7)
    assert rec.subject_data.gender == Gender.MALE
    assert rec.subject_data.age == 44
    assert rec.subject_data.weight == 89



# %%
from sleepecg import load_classifier, plot_hypnogram, read_slpdb, stage

# The model was built using tensorflow 2.7, running on higher versions might create warnings
# but should not influence the results.
clf = load_classifier("ws-gru-mesa", "SleepECG")

# %% Load record
# `ws-gru-mesa` performs poorly for most SLPDB records. It does however work well for slp03.
rec = next(read_slpdb("slp03"))

# %% Predict stages and plot hypnogram
stages_pred = stage(clf, rec, return_mode="prob")

plot_hypnogram(
    rec,
    stages_pred,
    stages_mode=clf.stages_mode,
    merge_annotations=True,
)



# © SleepECG developers
#
# License: BSD (3-clause)

"""Interface for listing and downloading NSRR (sleepdata.org) data."""

from __future__ import annotations

from fnmatch import fnmatch
from json.decoder import JSONDecodeError
from pathlib import Path

import requests
from tqdm import tqdm

from .utils import _download_file

_nsrr_token = None


def set_nsrr_token(token: str) -> None:
    """
    Set and verify the [NSRR](https://sleepdata.org) download token.

    Implemented according to the
    [NSRR API specs](https://github.com/nsrr/sleepdata.org/wiki/api-v1-account).

    Parameters
    ----------
    token : str
        NSRR [download token](https://sleepdata.org/token).
    """
    response = requests.get(
        "https://sleepdata.org/api/v1/account/profile.json",
        params={"auth_token": token},
    )
    authenticated = response.json()["authenticated"]
    if authenticated:
        username = response.json()["username"]
        email = response.json()["email"]
        print(f"Authenticated at sleepdata.org as {username} ({email})")
        global _nsrr_token
        _nsrr_token = token
    else:
        raise RuntimeError("Authentication at sleepdata.org failed, verify token!")


def _get_nsrr_url(db_slug: str) -> str:
    """
    Get the download URL for a given NSRR database.

    The download token is a part of the URL, so it needs to be already set.

    Parameters
    ----------
    db_slug : str
        Short identifier of a database, e.g. `'mesa'`.

    Returns
    -------
    str
        The download URL.
    """
    if _nsrr_token is None:
        raise RuntimeError("NSRR token not set, use `sleepecg.set_nsrr_token(<token>)`!")
    return f"https://sleepdata.org/datasets/{db_slug}/files/a/{_nsrr_token}/m/sleepecg/"


def _list_nsrr(
    db_slug: str,
    subfolder: str = "",
    pattern: str = "*",
    shallow: bool = False,
) -> list[tuple[str, str]]:
    """
    Recursively list filenames and checksums for a dataset.

    Specify a subfolder and/or a filename-pattern to filter results.

    Implemented according to the NSRR API documentation:
    https://github.com/nsrr/sleepdata.org/wiki/api-v1-datasets#list-files-in-folder

    Parameters
    ----------
    db_slug : str
        Short identifier of a database, e.g. `'mesa'`.
    subfolder : str, optional
        The folder at which to start the search, by default `''` (i.e. the root folder).
    pattern : str, optional
        Glob-like pattern to select files (only applied to the basename, not the dirname),
        by default `'*'`.
    shallow : bool, optional
        If `True`, only search in the given subfolder (i.e. no recursion), by default
        `False`.

    Returns
    -------
    list[tuple[str, str]]
        A list of tuples `(<filename>, <checksum>)`; `<filename>` is the full filename (i.e.
        dirname and basename) and `<checksum>` the MD5 checksum.
    """
    api_url = f"https://sleepdata.org/api/v1/datasets/{db_slug}/files.json"

    response = requests.get(api_url, params={"path": subfolder})
    try:
        response_json = response.json()
    except JSONDecodeError:
        raise RuntimeError(f"No API response for dataset {db_slug}.") from None

    files = []
    for item in response_json:
        if not item["is_file"] and not shallow:
            files.extend(_list_nsrr(db_slug, item["full_path"], pattern))
        elif fnmatch(item["file_name"], pattern):
            files.append((item["full_path"], item["file_checksum_md5"]))
    return files


def _download_nsrr_file(
    url: str,
    target_filepath: Path,
    checksum: str,
) -> None:
    """
    Download a file from `url` to `target_filepath` and verify `checksum`.

    This is a wrapper around `sleepecg.io.utils._download_file` to provide a helpful error
    message in case the currently set token does not grant access to the requested file.

    Parameters
    ----------
    url : str
        URL to download from.
    target_filepath : pathlib.Path
        Location where the downloaded file will be stored.
    checksum : str
        Checksum to verify the file against.
    """
    try:
        _download_file(url, target_filepath, checksum, "md5")
    except RuntimeError as error:
        # If the token is invalid for the requested dataset, the request is redirected to a
        # files overview page. The response is an HTML-page which doesn't have a
        # "content-disposition" header.
        response = requests.get(url, stream=True)
        if "content-disposition" not in response.headers:
            db_slug = url.split("/")[4]
            raise RuntimeError(f"Make sure you have access to {db_slug}!") from error
        else:
            raise


def download_nsrr(
    db_slug: str,
    subfolder: str = "",
    pattern: str = "*",
    shallow: bool = False,
    data_dir: str | Path = ".",
) -> None:
    """
    Recursively download files from [NSRR](https://sleepdata.org).

    Specify a subfolder and/or a filename pattern to filter results. Implemented according
    to the [NSRR API specs](https://github.com/nsrr/sleepdata.org/wiki/api-v1-datasets).

    Parameters
    ----------
    db_slug : str
        Short identifier of a database, e.g. `'mesa'`.
    subfolder : str, optional
        The folder at which to start the search, by default `''` (i.e. the root folder).
    pattern : str, optional
        Glob-like pattern to select files (only applied to the basename, not the dirname),
        by default `'*'`.
    shallow : bool, optional
        If `True`, only download files in the given subfolder (i.e. no recursion), by
        default `False`.
    data_dir : str | pathlib.Path, optional
        Directory where all datasets are stored, by default `'.'`.
    """
    db_dir = Path(data_dir) / db_slug

    download_url = _get_nsrr_url(db_slug)
    files_to_download = _list_nsrr(db_slug, subfolder, pattern, shallow)
    tqdm_description = f"Downloading {db_slug}/{subfolder or '.'}/{pattern}"

    for filepath, checksum in tqdm(files_to_download, desc=tqdm_description):
        target_filepath = db_dir / filepath
        url = download_url + filepath
        _download_nsrr_file(url, target_filepath, checksum)
